@@ -51,12 +51,21 @@ class Pair(object):
     chain, formatted as would be returned from a query on a MongoDB database containing
     AbStar output.
     '''
-    def __init__(self, seqs, name=None, h_selection_func=None, l_selection_func=None):
+    def __init__(self, seqs, name=None, chain_selection_func=None):
         self._seqs = seqs
+        self._receptor = None
         self._heavy = None
         self._light = None
-        self._heavies = [s for s in seqs if s['chain'] == 'heavy']
-        self._lights = [s for s in seqs if s['chain'] in ['kappa', 'lambda']]
+        self._alpha = None
+        self._beta = None
+        self._delta = None
+        self._gamma = None
+        self._heavies = None
+        self._lights = None
+        self._alphas = None
+        self._betas = None
+        self._deltas = None
+        self._gammas = None
         self._name = name
         self._fasta = None
         self._sample = None
@@ -65,11 +74,9 @@ class Pair(object):
         self._experiment = None
         self._timepoint = None
         self._is_pair = None
-        # self._vrc01_like = None
-        # self._loose_vrc01_like = None
         self._lineage = None
-        self._select_heavy = h_selection_func
-        self._select_light = l_selection_func
+        self._select_chain = chain_selection_func if chain_selection_func is not None else self._chain_selector
+
 
     def __eq__(self, other):
         return (self.heavy, self.light) == (other.heavy, other.light)
@@ -77,26 +84,30 @@ class Pair(object):
     def __ne__(self, other):
         return not self == other
 
-    def __hash(self):
+    def __hash__(self):
         return hash((self.heavy, self.light))
+        
+
+    @property
+    def receptor(self):
+        if self._receptor is None:
+            if all([s['chain'] in ['heavy', 'kappa', 'lambda'] for s in self._seqs]):
+                self._receptor = 'bcr'
+            elif all([s['chain'] in ['alpha', 'beta', 'delta', 'gamma'] for s in self._seqs]):
+                self._receptor = 'tcr'
+            else:
+                self._receptor = 'unknown'
+        return self._receptor
+
+    @receptor.setter
+    def receptor(self, receptor):
+        self._receptor = receptor
 
 
     @property
     def heavy(self):
         if self._heavy is None:
-            if len(self._heavies) > 0:
-                if self._select_heavy is not None:
-                    h = self._select_heavy(self._heavies)
-                    if all([h is not None, type(h) != Sequence]):
-                        h = Sequence(h)
-                    self._heavy = h
-                else:
-                    h = self._heavies[0]
-                    if type(h) != Sequence:
-                        h = Sequence(h)
-                    self._heavy = h
-            else:
-                self._heavy = None
+            self._heavy = self._select_chain(self.heavies)
         return self._heavy
 
     @heavy.setter
@@ -106,38 +117,120 @@ class Pair(object):
     @property
     def light(self):
         if self._light is None:
-            if len(self._lights) > 0:
-                if self._select_light is not None:
-                    l = self._select_light(self._lights)
-                    if all([l is not None, type(l) != Sequence]):
-                        l = Sequence(l)
-                    self._light = l
-                else:
-                    l = self._lights[0]
-                    if type(l) != Sequence:
-                        l = Sequence(l)
-                    self._light = l
-            else:
-                self._light = None
+            self._light = self._select_chain(self.lights)
         return self._light
 
     @light.setter
     def light(self, light):
         self._light = light
 
+
+    @property
+    def alpha(self):
+        if self._alpha is None:
+            self._alpha = self._select_chain(self.alphas)
+        return self._alpha
+
+    @alpha.setter
+    def alpha(self, alpha):
+        self._alpha = alpha
+
+    
+    @property
+    def beta(self):
+        if self._beta is None:
+            self._beta = self._select_chain(self.betas)
+        return self._beta
+
+    @beta.setter
+    def beta(self, beta):
+        self._beta = beta
+
+    
+    @property
+    def delta(self):
+        if self._delta is None:
+            self._delta = self._select_chain(self.deltas)
+        return self._delta
+
+    @delta.setter
+    def delta(self, delta):
+        self._delta = delta
+
+
+    @property
+    def gamma(self):
+        if self._gamma is None:
+            self._gamma = self._select_chain(self.gammas)
+        return self._gamma
+
+    @gamma.setter
+    def gamma(self, gamma):
+        self._gamma = gamma
+
     
     @property
     def heavies(self):
+        if self._heavies is None:
+            if all([s['chain'] is not None for s in self._seqs]):
+                self._heavies = [s for s in self._seqs if s['chain'] == 'heavy']
+            elif all([s['locus'] is not None for s in self._seqs]):
+                self._heavies = [s for s in self._seqs if s['locus'] == 'IGH']
         return self._heavies
     
     @property
     def lights(self):
+        if self._lights is None:
+            if all([s['chain'] is not None for s in self._seqs]):
+                self._lights = [s for s in self._seqs if s['chain'] in ['kappa', 'lambda']]
+            elif all([s['locus'] is not None for s in self._seqs]):
+                self._lights = [s for s in self._seqs if s['locus'] in ['IGK', 'IGL']]
         return self._lights
+    
+    @property
+    def alphas(self):
+        if self._alphas is None:
+            if all([s['chain'] is not None for s in self._seqs]):
+                self._alphas = [s for s in self._seqs if s['chain'] == 'alpha']
+            elif all([s['locus'] is not None for s in self._seqs]):
+                self._alphas = [s for s in self._seqs if s['locus'] == 'TRA']
+        return self._alphas
+
+    @property
+    def betas(self):
+        if self._betas is None:
+            if all([s['chain'] is not None for s in self._seqs]):
+                self._betas = [s for s in self._seqs if s['chain'] == 'beta']
+            elif all([s['locus'] is not None for s in self._seqs]):
+                self._betas = [s for s in self._seqs if s['locus'] == 'TRB']
+        return self._betas
+
+    @property
+    def deltas(self):
+        if self._deltas is None:
+            if all([s['chain'] is not None for s in self._seqs]):
+                self._deltas = [s for s in self._seqs if s['chain'] == 'delta']
+            elif all([s['locus'] is not None for s in self._seqs]):
+                self._deltas = [s for s in self._seqs if s['locus'] == 'TRD']
+        return self._deltas
+
+    @property
+    def gammas(self):
+        if self._gammas is None:
+            if all([s['chain'] is not None for s in self._seqs]):
+                self._gammas = [s for s in self._seqs if s['chain'] == 'gamma']
+            elif all([s['locus'] is not None for s in self._seqs]):
+                self._gammas = [s for s in self._seqs if s['locus'] == 'TRG']
+        return self._gammas
 
     @property
     def is_pair(self):
         if all([self.heavy is not None, self.light is not None]):
             return True
+        elif all([self.alpha is not None, self.beta is not None]):
+            return True
+        elif all([self.gamma is not None, self.delta is not None]):
+            return True
         return False
 
     @property
@@ -361,7 +454,7 @@ class Pair(object):
     #     seq['vdj_aa'] = Seq(seq['vdj_nt']).translate()
 
 
-    def fasta(self, name_field='seq_id', sequence_field='vdj_nt', append_chain=True):
+    def fasta(self, name_field='sequence_id', sequence_field='sequence', append_chain=True):
         '''
         Returns the sequence pair as a fasta string. If the Pair object contains
         both heavy and light chain sequences, both will be returned as a single string.
@@ -384,8 +477,23 @@ class Pair(object):
         return '\n'.join(fastas)
 
 
+    def _chain_selector(self, seqs):
+        if len(seqs) == 0:
+            return None
+        if all(['umis' in s for s in seqs]):
+            sorted_seqs = sorted(seqs, key=lambda x: x['umis'], reverse=True)
+            return sorted_seqs[0]
+        else:
+            return seqs[0]
+
+
+
+
+
+
+
 def get_pairs(db, collection, experiment=None, subject=None, group=None, name='seq_id',
-    delim=None, delim_occurance=1, pairs_only=False, h_selection_func=None, l_selection_func=None):
+    delim=None, delim_occurance=1, pairs_only=False, chain_selection_func=None):
     '''
     Gets sequences and assigns them to the appropriate mAb pair, based on the sequence name.
 
@@ -430,11 +538,11 @@ def get_pairs(db, collection, experiment=None, subject=None, group=None, name='s
     seqs = list(db[collection].find(match))
     return assign_pairs(seqs, name=name, delim=delim,
         delim_occurance=delim_occurance, pairs_only=pairs_only,
-        h_selection_func=h_selection_func, l_selection_func=l_selection_func)
+        chain_selection_func=chain_selection_func)
 
 
-def assign_pairs(seqs, name='seq_id', delim=None, delim_occurance=1, pairs_only=False,
-                 h_selection_func=None, l_selection_func=None, tenx_annot_file=None):
+def assign_pairs(seqs, id_key='sequence_id', delim=None, delim_occurance=1, pairs_only=False,
+                 chain_selection_func=None, tenx_annot_file=None):
     '''
     Assigns sequences to the appropriate mAb pair, based on the sequence name.
 
@@ -457,96 +565,91 @@ def assign_pairs(seqs, name='seq_id', delim=None, delim_occurance=1, pairs_only=
     # add UMIs to the sequence annotations if a 10xG annotations file is provided
     if tenx_annot_file is not None:
         annots = {}
+        umis = {}
         with open(tenx_annot_file) as f:
             reader = csv.DictReader(f)
             for r in reader:
                 key = 'consensus_id' if 'consensus_id' in r else 'contig_id'
-                annots[r[key]] = r['umis']
+                umis[r[key]] = r['umis']
+                annots[r[key]] = r
         for s in seqs:
-            s['umis'] = annots.get(s['seq_id'], 0)
-        if h_selection_func is None:
-            h_selection_func = umi_selector
-        if l_selection_func is None:
-            l_selection_func = umi_selector
+            s['umis'] = umis.get(s[id_key], 0)
+            s['tenx_annots'] = annots.get(s[id_key], {})
+
     # identify pairs
     pdict = {}
     for s in seqs:
         if delim is not None:
-            pname = delim.join(s[name].split(delim)[:delim_occurance])
+            pname = delim.join(s[id_key].split(delim)[:delim_occurance])
         else:
-            pname = s[name]
+            pname = s[id_key]
         if pname not in pdict:
             pdict[pname] = [s, ]
         else:
             pdict[pname].append(s)
-    pairs = [Pair(pdict[n], name=n,
-                  h_selection_func=h_selection_func,
-                  l_selection_func=l_selection_func) for n in pdict.keys()]
+    pairs = [Pair(p,
+                  name=n,
+                  chain_selection_func=chain_selection_func) for n, p in pdict.items()]
     if pairs_only:
         pairs = [p for p in pairs if p.is_pair]
     return pairs
 
 
-def umi_selector(seqs):
-    sorted_seqs = sorted(seqs, key=lambda x: x['umis'], reverse=True)
-    return sorted_seqs[0]
+# def deduplicate(pairs, aa=False, ignore_primer_regions=False):
+#     '''
+#     Removes duplicate sequences from a list of Pair objects.
+
+#     If a Pair has heavy and light chains, both chains must identically match heavy and light chains
+#     from another Pair to be considered a duplicate. If a Pair has only a single chain,
+#     identical matches to that chain will cause the single chain Pair to be considered a duplicate,
+#     even if the comparison Pair has both chains.
+
+#     Note that identical sequences are identified by simple string comparison, so sequences of
+#     different length that are identical over the entirety of the shorter sequence are not
+#     considered duplicates.
+
+#     By default, comparison is made on the nucleotide sequence. To use the amino acid sequence instead,
+#     set aa=True.
+#     '''
+#     nr_pairs = []
+#     just_pairs = [p for p in pairs if p.is_pair]
+#     single_chains = [p for p in pairs if not p.is_pair]
+#     _pairs = just_pairs + single_chains
+#     for p in _pairs:
+#         duplicates = []
+#         for nr in nr_pairs:
+#             identical = True
+#             vdj = 'vdj_aa' if aa else 'vdj_nt'
+#             offset = 4 if aa else 12
+#             if p.heavy is not None:
+#                 if nr.heavy is None:
+#                     identical = False
+#                 else:
+#                     heavy = p.heavy[vdj][offset:-offset] if ignore_primer_regions else p.heavy[vdj]
+#                     nr_heavy = nr.heavy[vdj][offset:-offset] if ignore_primer_regions else nr.heavy[vdj]
+#                     if heavy != nr_heavy:
+#                         identical = False
+#             if p.light is not None:
+#                 if nr.light is None:
+#                     identical = False
+#                 else:
+#                     light = p.light[vdj][offset:-offset] if ignore_primer_regions else p.light[vdj]
+#                     nr_light = nr.light[vdj][offset:-offset] if ignore_primer_regions else nr.light[vdj]
+#                     if light != nr_light:
+#                         identical = False
+#             duplicates.append(identical)
+#         if any(duplicates):
+#             continue
+#         else:
+#             nr_pairs.append(p)
+#     return nr_pairs
 
 
-def deduplicate(pairs, aa=False, ignore_primer_regions=False):
-    '''
-    Removes duplicate sequences from a list of Pair objects.
-
-    If a Pair has heavy and light chains, both chains must identically match heavy and light chains
-    from another Pair to be considered a duplicate. If a Pair has only a single chain,
-    identical matches to that chain will cause the single chain Pair to be considered a duplicate,
-    even if the comparison Pair has both chains.
-
-    Note that identical sequences are identified by simple string comparison, so sequences of
-    different length that are identical over the entirety of the shorter sequence are not
-    considered duplicates.
-
-    By default, comparison is made on the nucleotide sequence. To use the amino acid sequence instead,
-    set aa=True.
-    '''
-    nr_pairs = []
-    just_pairs = [p for p in pairs if p.is_pair]
-    single_chains = [p for p in pairs if not p.is_pair]
-    _pairs = just_pairs + single_chains
-    for p in _pairs:
-        duplicates = []
-        for nr in nr_pairs:
-            identical = True
-            vdj = 'vdj_aa' if aa else 'vdj_nt'
-            offset = 4 if aa else 12
-            if p.heavy is not None:
-                if nr.heavy is None:
-                    identical = False
-                else:
-                    heavy = p.heavy[vdj][offset:-offset] if ignore_primer_regions else p.heavy[vdj]
-                    nr_heavy = nr.heavy[vdj][offset:-offset] if ignore_primer_regions else nr.heavy[vdj]
-                    if heavy != nr_heavy:
-                        identical = False
-            if p.light is not None:
-                if nr.light is None:
-                    identical = False
-                else:
-                    light = p.light[vdj][offset:-offset] if ignore_primer_regions else p.light[vdj]
-                    nr_light = nr.light[vdj][offset:-offset] if ignore_primer_regions else nr.light[vdj]
-                    if light != nr_light:
-                        identical = False
-            duplicates.append(identical)
-        if any(duplicates):
-            continue
-        else:
-            nr_pairs.append(p)
-    return nr_pairs
-
-
-def refine(pairs, heavy=True, light=True, species='human'):
-    refined_pairs = copy.deepcopy(pairs)
-    for p in refined_pairs:
-        p.refine(heavy, light, species)
-    return refined_pairs
+# def refine(pairs, heavy=True, light=True, species='human'):
+#     refined_pairs = copy.deepcopy(pairs)
+#     for p in refined_pairs:
+#         p.refine(heavy, light, species)
+#     return refined_pairs
 
 
 
